from __future__ import annotations

from collections.abc import Iterable
from decimal import Decimal
from typing import IO, Any

import jsonlines
import orjson
from geoalchemy2 import functions as func
from sqlalchemy import Column, MetaData, select
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import ClauseElement

from schematools.exports import BaseExporter
from schematools.exports.csv import DatasetFieldSchema, enable_datetime_cast
from schematools.naming import toCamelCase
from schematools.types import DatasetSchema, DatasetTableSchema

metadata = MetaData()


def _default(obj: Any) -> str:
    if isinstance(obj, Decimal):
        return str(obj)
    raise TypeError


def _dumps(obj: Any) -> str:
    """Json serializer.

    Unfortunately, orjson does not support Decimal serialization,
    so we need this extra function.
    """
    return orjson.dumps(obj, default=_default)


class JsonLinesExporter(BaseExporter):  # noqa: D101
    extension = "jsonl"

    def geo_modifier(field: DatasetFieldSchema, column):
        if not field.is_geo:
            return column
        return func.ST_AsGeoJSON(func.ST_Transform(column, 4326)).label(field.db_name)

    def id_modifier(field: DatasetFieldSchema, column):
        if field.table.is_temporal and field.is_composite_key:
            return func.split_part(column, ".", 1).label(field.db_name)
        return column

    # We do not use the iso for datetime here, because json notation handles this.

    processors = (geo_modifier, id_modifier)

    def _get_row_modifier(self, table: DatasetTableSchema):
        lookup = {}
        for field in table.fields:
            lookup[field.db_name] = (
                (lambda v: orjson.loads(v) if v else v)
                if field.is_geo or field.is_nested_object
                else lambda v: v
            )
        return lookup

    def write_rows(  # noqa: D102
        self,
        file_handle: IO[str],
        table: DatasetTableSchema,
        columns: Iterable[Column],
        temporal_clause: ClauseElement | None,
        srid: str,
    ):
        writer = jsonlines.Writer(file_handle, dumps=_dumps)
        row_modifier = self._get_row_modifier(table)
        query = select(*columns)
        if temporal_clause is not None:
            query = query.where(temporal_clause)
        if self.size is not None:
            query = query.limit(self.size)

        with self.connection.execution_options(yield_per=1000).connect() as conn:
            result = conn.execute(query)
            for partition in result.mappings().partitions():
                for r in partition:
                    writer.write({toCamelCase(k): row_modifier[k](v) for k, v in r.items()})


def export_jsonls(
    connection: Connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
):
    """Utility function to wrap the Exporter."""
    enable_datetime_cast()
    exporter = JsonLinesExporter(connection, dataset_schema, output, table_ids, scopes, size)
    exporter.export_tables()



from __future__ import annotations

from django.db import models


class UnlimitedCharField(models.CharField):
    # When we switch to Django 4.2, replace this with CharField(max_length=None).
    description = "CharField without length limit"
    max_length = (1 << 31) - 1  # Big limit to fool validation code.

    def __init__(self, *args, **kwargs):
        if "max_length" in kwargs:
            kwargs.pop("max_length")
        super().__init__(*args, max_length=self.max_length, **kwargs)

    def db_type(self, connection):
        return "varchar"



from __future__ import annotations

import json
import re
from collections.abc import Iterable
from typing import Any

from shapely.geometry import shape

from schematools.exceptions import ParserError
from schematools.types import DatasetTableSchema

from .base import BaseImporter, Record

ID_FORMAT = re.compile(r"^([a-z0-9_]+)[/.](\d+)$", re.I)


def read_geojson(file_name) -> Iterable[dict]:
    """Read the contents of a GeoJSON file into memory.
    This also performs basic checks whether the data is valid.
    """
    with open(file_name) as fh:
        geojson = json.load(fh)
        if (
            not isinstance(geojson, dict)
            or geojson.get("type") != "FeatureCollection"
            or not isinstance(geojson.get("features"), list)
        ):
            raise ParserError(f"{file_name} is not a valid GeoJSON file")

    for feature in geojson["features"]:
        feature_type = feature.get("type")
        if feature_type != "Feature":
            raise ParserError(f"Expected 'Feature' in {file_name}, not {feature_type}")

        yield feature


def split_id(id_value) -> tuple[str, str]:
    # When the ID format is name/identifier,
    # this detects that different feature types are part of the same file.
    match = ID_FORMAT.match(id_value)
    if match:
        return match.group(1), match.group(2)
    else:
        raise ValueError("Can't split ID value")


class GeoJSONImporter(BaseImporter):
    """Import an GeoJSON file into the database."""

    def parse_records(self, file_name, dataset_table: DatasetTableSchema, **kwargs):
        """Provide an iterator the reads the NDJSON records"""
        features = read_geojson(file_name)
        main_geometry = dataset_table.main_geometry_field

        relation_fields = [field for field in dataset_table.fields if field.relation is not None]
        for feature in features:
            record = Record(
                (self._clean_value(name, value) for name, value in feature["properties"].items()),
                source=feature,
            )
            for relation_field in relation_fields:
                if relation_field.db_name != relation_field.python_name:
                    record[relation_field.db_name] = record.pop(relation_field.python_name)

            wkt = shape(feature["geometry"]).wkt
            record[main_geometry.db_name] = f"SRID={main_geometry.srid};{wkt}"

            yield {dataset_table.id: [record]}

    def _clean_value(self, name: str, value: Any) -> tuple[str, Any]:
        if name[0] in "@$":
            name = name[1:]

        if name == "id":
            try:
                value = split_id(value)[1]
            except ValueError:
                pass

        return name, value



from __future__ import annotations

import json
import re
from pathlib import Path

import pytest

from schematools import validation
from schematools.permissions import PUBLIC_SCOPE
from schematools.types import DatasetSchema
from schematools.validation import (
    PROPERTIES_INTRODUCING_BREAKING_CHANGES,
    _active_versions,
    _check_display,
    _check_maingeometry,
    _identifier_properties,
    validate_table,
)


def test_camelcase() -> None:
    for ident in ("camelCase", "camelCase100"):
        assert validation._camelcase_ident(ident) is None

    error = validation._camelcase_ident("")
    assert error is not None
    assert "empty identifier" in error

    for ident, suggest in (
        ("snake_case", "snakeCase"),
        ("camelCase_snake", "camelCaseSnake"),
        ("camel100camel", "camel100Camel"),
    ):
        error = validation._camelcase_ident(ident)
        assert error is not None
        assert error.endswith(f"suggestion: {suggest}")


def test_enum_types(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("enum_types.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "value 'foo' in field enumInts is not an integer"

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "value 2 in field enumStrs is not a string"

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "enumFloats: enum of type number not possible"

    assert list(errors) == []


def test_id_auth(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("id_auth.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "Auth on identifier field"
    assert """auth on field 'id'""" in error.message

    assert list(errors) == []


def test_id_type(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("id_type.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "Identifier field with the wrong type"
    assert """field 'uniqid'""" in error.message
    assert """'number'""" in error.message

    assert list(errors) == []


def test_id_matches_path(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")

    # No errors when id equals parent path name
    errors = validation.run(dataset, str(here / "files/datasets/stadsdelen/dataset.json"))
    assert list(errors) == []

    # Error when not equal
    errors = validation.run(dataset, str(here / "files/datasets/regios/dataset.json"))
    error = next(errors)
    assert error
    assert error.validator_name == "ID does not match file path"
    assert re.match(
        r"^Id of the dataset stadsdelen does not match the parent directory"
        r" .*/files/datasets/regios\.$",
        error.message,
    )
    assert list(errors) == []

    # Test datasets in sub directory
    dataset.__setitem__("id", "beheerkaartCbsGrid")
    errors = validation.run(dataset, str(here / "files/datasets/bierkaart/cbs_grid/dataset.json"))
    error = next(errors)
    assert error
    assert error.validator_name == "ID does not match file path"
    assert re.match(
        r"^Id of the dataset beheerkaartCbsGrid does not match the parent directory"
        r" .*/files/datasets/bierkaart/cbs_grid\.$",
        error.message,
    )
    assert list(errors) == []

    errors = validation.run(
        dataset, str(here / "files/datasets/beheerkaart/cbs_grid/dataset.json")
    )
    assert list(errors) == []

    # Test identifiers ending with a number
    dataset.__setitem__("id", "covid19")
    errors = validation.run(dataset, str(here / "files/datasets/covid19/dataset.json"))
    assert list(errors) == []


def test_crs(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("crs_validation.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "crs"
    assert """No coordinate reference system defined for field""" in error.message

    assert list(errors) == []


def test_postgres_identifier_length(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("long_ids.json")

    error = next(validation.run(dataset))
    assert error
    assert error.validator_name == "PostgreSQL identifier length"
    assert "absurdly_long" in error.message

    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")
    assert list(validation.run(dataset)) == []  # no validation errors


def test_identifier_properties(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("identifier_ref.json")
    error = next(validation.run(dataset))
    assert error
    assert "foobar" in error.message

    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")
    assert list(_identifier_properties(dataset)) == []  # no validation errors


def test_active_versions(schema_loader) -> None:
    dataset = schema_loader.get_dataset("gebieden_sep_tables")
    table_versions = dataset.table_versions["bouwblokken"]
    table_versions.id = "BOUWBLOKKEN"
    error = next(validation.run(dataset))
    assert error
    assert "does not match with id" in error.message

    dataset = schema_loader.get_dataset("gebieden_sep_tables")
    dataset["tables"][0]["activeVersions"] = {"9.8.1": "bouwblokken/v1.0.0"}
    error = next(validation.run(dataset))
    assert error
    assert "does not match with version" in error.message

    dataset = schema_loader.get_dataset("gebieden_sep_tables")
    assert list(_active_versions(dataset)) == []  # no validation errors


def test_main_geometry(schema_loader, gebieden_schema) -> None:
    dataset = schema_loader.get_dataset_from_file("meetbouten.json")
    assert list(_check_maingeometry(dataset)) == []

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = None
    error = next(validation.run(dataset))
    assert "'mainGeometry' is required but not defined in table" in error.message

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = "not_a_geometry"
    error = next(validation.run(dataset))
    assert "mainGeometry = 'not_a_geometry'" in error.message
    assert "Field 'not_a_geometry' does not exist" in error.message

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = "merkOmschrijving"
    error = next(validation.run(dataset))
    assert error.message == (
        "mainGeometry = 'merkOmschrijving' is not a geometry field, type = 'string'"
    )


def test_display(here: Path, schema_loader) -> None:
    schema_loader.get_dataset_from_file("gebieden.json")  # fill cache
    dataset = schema_loader.get_dataset_from_file("meetbouten.json")
    assert list(_check_display(dataset)) == []

    table = dataset.get_table_by_id("meetbouten")
    table["schema"]["display"] = "not_a_field"
    error = next(validation.run(dataset))
    assert "display = 'not_a_field'" in error.message
    assert "Field 'not_a_field' does not exist" in error.message

    table["schema"]["display"] = "merkCode"
    table["schema"]["properties"]["merkCode"]["auth"] = "some_scope"
    table.__dict__.pop("fields", None)  # clear cached property
    error = next(validation.run(dataset))
    assert "'auth' property on the display field: 'merkCode' is not allowed." in error.message


def test_rel_auth_dataset(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("rel_auth.json")
    dataset["auth"] = ["HAMMERTIME"]
    dataset["reasonsNonPublic"] = ["U can't touch this"]

    errors = list(validation.run(dataset))
    assert errors == []


def test_rel_auth_dataset_public(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("rel_auth.json")
    dataset["auth"] = [PUBLIC_SCOPE]

    errors = list(validation.run(dataset))
    assert len(errors) == 0, errors


def test_rel_auth_table(here: Path) -> None:
    with (here / "files/datasets/rel_auth.json").open() as f:
        dataset_json = json.load(f)
    table = next(t for t in dataset_json["tables"] if t["id"] == "base")
    table["auth"] = ["HAMMERTIME"]
    table["reasonsNonPublic"] = ["U can't touch this"]
    dataset = DatasetSchema.from_dict(dataset_json)

    errors = list(validation.run(dataset))
    assert len(errors) == 1, errors
    assert "requires scopes ['HAMMERTIME']" in str(errors[0])


def test_rel_auth_field(here: Path) -> None:
    with (here / "files/datasets/rel_auth.json").open() as f:
        dataset_json = json.load(f)
    table = next(t for t in dataset_json["tables"] if t["id"] == "base")
    field = table["schema"]["properties"]["stop"]
    field["auth"] = ["HAMMERTIME"]

    dataset = DatasetSchema.from_dict(dataset_json)
    errors = list(validation.run(dataset))

    assert len(errors) >= 1, errors
    assert any("requires scopes ['HAMMERTIME']" in str(e) for e in errors)


@pytest.mark.skip(reason="See comment in validator function")
def test_repetitive_naming(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset("repetitive")
    errors = {str(e) for e in validation.run(dataset)}

    assert errors == {
        "[repetitive identifiers] " + e
        for e in [
            "table name 'repetitiveTable' should not start with 'repetitive'",
            "field name 'repetitiveTableField' should not start with 'repetitive'",
            "field name 'repetitiveTableField' should not start with 'repetitiveTable'",
        ]
    }


def test_reasons_non_public_exists(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("hr_auth.json")
    errors = list(validation.run(dataset))

    # Test an error is given for the highest non-public scope
    # and only for the highest non-public scope.
    assert len(errors) == 1
    assert errors[0].message == "Non-public dataset hr should have a 'reasonsNonPublic' property."

    dataset["auth"] = [PUBLIC_SCOPE]
    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        errors[0].message
        == "Non-public table sbiactiviteiten should have a 'reasonsNonPublic' property."
    )

    # Test no error is given when a reason is present
    dataset_json = json.load((here / "files/datasets/hr_auth.json").open())
    dataset_json["tables"][0]["reasonsNonPublic"] = ["5.1 1c: Bevat persoonsgegevens"]
    dataset = DatasetSchema.from_dict(dataset_json)
    dataset["auth"] = [PUBLIC_SCOPE]
    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_reasons_non_public_value(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("hr_auth.json")
    dataset["reasonsNonPublic"] = ["5.1 1c: Bevat persoonsgegevens", "nader te bepalen"]
    errors = list(validation.run(dataset))

    # Test an error is given for the placeholder value in a dataset with status = beschikbaar.
    assert len(errors) == 1
    assert "not allowed in ReasonsNonPublic property of dataset hr." in errors[0].message

    # Test no error is given for the placeholder value in a dataset with status != beschikbaar.
    dataset["status"] = "niet_beschikbaar"
    errors = list(validation.run(dataset))
    assert len(errors) == 0

    # Test no error is given for other values of reasonsNonPublic
    dataset["status"] = "beschikbaar"
    dataset["reasonsNonPublic"] = ["5.1 1c: Bevat persoonsgegevens"]
    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_schema_ref(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("schema_ref_validation.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "schema ref"
    assert """Incorrect `$ref` for""" in error.message

    assert list(errors) == []


def test_check_default_version(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("schema_default_version.json")

    errors = list(validation.run(dataset))
    assert len(errors) == 0

    # Prove that wrong default version gives an error
    dataset["defaultVersion"] = "v2"
    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert "Default version v2 does not match enabled version v1" in errors[0].message


def test_production_version_tables(schema_loader) -> None:
    dataset = schema_loader.get_dataset("production_version")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Dataset version (v1) cannot contain non-production table [tables/v0]" in errors[0].message
    )


def test_production_version_experimental_tables(schema_loader) -> None:
    dataset = schema_loader.get_dataset("experimental_tables")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Stable dataset experimental_tables (v1) cannot have tables with lifecycleStatus of 'experimental'."
        in errors[0].message
    )


def test_check_lifecycle_status(schema_loader) -> None:
    dataset = schema_loader.get_dataset("lifecycle_status")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Dataset version (v0) cannot have a lifecycleStatus of 'stable' while being a non-production version."
        in errors[0].message
    )


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        # No changes
        ({"name": {"type": "string"}}, {"name": {"type": "string"}}, []),
        # Deleted field
        ({"name": {"type": "string"}}, {}, ["Column name would be deleted."]),
        # Changed array item type
        (
            {"list": {"type": "array", "items": {"type": "string"}}},
            {"list": {"type": "array", "items": {"type": "integer"}}},
            ["Column list would change items.type."],
        ),
        # Changed object property.
        (
            {"object": {"type": "object", "properties": {"element": {"type": "string"}}}},
            {"object": {"type": "object", "properties": {"element": {"type": "integer"}}}},
            ["Column object would change element.type."],
        ),
        # Changed object property type within an array
        (
            {
                "list": {
                    "type": "array",
                    "items": {"type": "object", "properties": {"element": {"type": "string"}}},
                }
            },
            {
                "list": {
                    "type": "array",
                    "items": {"type": "object", "properties": {"element": {"type": "integer"}}},
                }
            },
            ["Column list would change items.element.type."],
        ),
        # Changed subproperty of a property of an object
        (
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "string"}},
                        }
                    },
                }
            },
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "integer"}},
                        }
                    },
                }
            },
            ["Column object would change element.subelement.type."],
        ),
        # Multiple breaking changes
        (
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {
                                "subelement": {"type": "string"},
                                "subelement2": {"type": "string"},
                            },
                        }
                    },
                    "description": "Original description",
                    "relation": "table1:object_1",
                },
                "id": {"type": "string"},
                "deprecated_field": {"type": "string"},
            },
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "integer"}},
                        }
                    },
                    "description": "Description 2.0",
                    "relation": "table2:object_2",
                },
                "id": {"type": "integer"},
            },
            [
                "Column object would change relation.",
                "Column object would change element.subelement.type.",
                "Property element.subelement2 would be deleted from column object.",
                "Column id would change type.",
                "Column deprecated_field would be deleted.",
            ],
        ),
    ]
    + [
        # All properties that should stay the same.
        (
            {"property": {prop: "string"}},
            {"property": {prop: "integer"}},
            [f"Column property would change {prop}."],
        )
        for prop in PROPERTIES_INTRODUCING_BREAKING_CHANGES
    ],
)
def test_validate_table(prev, curr, errors):
    table_errors = validate_table(prev, curr)
    assert table_errors == errors



"""Semantic JSON schema validation.

This package provides a simple :class:`Validator` class
that allows us to perform semantic validation
on a given JSON schema.

Semantic validation is different from structural validation,
as performed by the :mod:`jsonschema`,
in that it takes into account the implied logic behind a JSON schema.
For instance,
we might indicate that what identifies a table
is a set of properties (columns).
In Amsterdam Schema this is done using the ``identifier`` property
that can take an array of property names.
Structural validation cannot verify that the properties specified in that array
actually do exist in the schema.
Semantic validation can and should.

Semantic validations are created decorating a function with `_validator`.
"""

from __future__ import annotations

import operator
import re
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from functools import partial, wraps
from pathlib import Path
from typing import cast
from urllib.parse import urlparse

from schematools import MAX_TABLE_NAME_LENGTH
from schematools.exceptions import SchemaObjectNotFound
from schematools.naming import to_snake_case, toCamelCase
from schematools.types import DatasetSchema, DatasetTableSchema, DatasetVersionSchema


@dataclass(frozen=True)
class ValidationError:
    """Capture validation errors."""

    validator_name: str
    message: str

    def __str__(self) -> str:
        return f"[{self.validator_name}] {self.message}"


_all: list[tuple[str, Callable[[DatasetSchema, str | None], Iterator[str]]]] = []


def run(dataset: DatasetSchema, location: str | None = None) -> Iterator[ValidationError]:
    r"""Run all registered validators.

    Yields:
        :class:`ValidationError`\s, if any.

    """  # noqa: W605
    for name, validator in _all:
        for msg in validator(dataset, location):
            yield ValidationError(validator_name=name, message=msg)


def _register_validator(name: str) -> Callable:
    """Marks a function as a validator and registers it with `run`.

    The function should yield strings describing the problem.
    `run` combines those strings with `name` into ValidationErrors.
    """
    if not name:
        raise ValueError("validator must have a name")

    def decorator(func: Callable[[DatasetSchema, str | None], Iterator[str]]) -> Callable:
        @wraps(func)
        def decorated(dataset: DatasetSchema, location: str | None = None) -> Iterator[str]:
            if func.__code__.co_argcount == 1:
                return func(dataset)
            else:
                return func(dataset, location)

        _all.append((name, decorated))
        return decorated

    return decorator


@_register_validator("camel case")
def _camelcase(dataset: DatasetSchema) -> Iterator[str]:
    """Checks that conversion to snake case and back leaves field identifiers unchanged."""
    for table in dataset.tables:
        for field in table.fields:
            error = _camelcase_ident(field.id)
            if error is not None:
                yield error


def _camelcase_ident(ident: str) -> str | None:
    if ident == "":
        return "empty identifier not allowed"
    camel = toCamelCase(to_snake_case(ident))
    if camel == ident:
        return None
    return f"{ident} does not survive conversion to snake case and back; suggestion: {camel}"


@_register_validator("enum type error")
def _enum_types(dataset: DatasetSchema, location: str | None) -> Iterator[str]:
    for table in dataset.tables:
        for field in table.fields:
            enum = field.get("enum")
            if not enum:
                continue

            if field.type == "integer":
                typ = int
                a = "an"
            elif field.type == "string":
                typ = str
                a = "a"
            else:
                yield f"{field.id}: enum of type {field.type} not possible"
                continue

            for v in enum:
                if not isinstance(v, typ):
                    yield f"value {v!r} in field {field.id} is not {a} {field.type}"


@_register_validator("ID does not match file path")
def _id_matches_path(dataset: DatasetSchema, location: str | None) -> Iterator[str]:
    """Dataset Identifiers should equal the parent paths to assure uniqueness.

    For datasets is subdirectories the path components should match the id like:
    'my/nested/location/dataset.json' -> 'myNestedLocation'

    Arguments:
        location: Location of the dataset file (relative to root or absolute)
    """
    if location is not None:
        path = Path(location)
        id_ = to_snake_case(dataset.id)

        # Ids are allowed to end with a number,
        # but the number should not be camelCased in the path.
        # So in this case the last instance of '_' is removed
        if id_.split("_")[-1].isdigit():
            id_ = "".join(id_.rsplit("_", 1))
        temp_path = path
        while len(id_):
            temp_path = temp_path.parent
            if not id_.endswith(temp_path.name):
                yield (
                    f"Id of the dataset {dataset.id} does not match "
                    f"the parent directory {path.parent}."
                )
                break
            id_ = id_[: -len(temp_path.name) - 1]


@_register_validator("Auth on identifier field")
def _id_auth(dataset: DatasetSchema) -> Iterator[str]:
    """Identifier fields should not have "auth" scopes.

    Handling these separately from table scopes is too much work for too little gain.
    """
    for table in dataset.tables:
        for ident in table.identifier:
            try:
                field = table.get_field_by_id(ident)
                if field.auth != {"OPENBAAR"}:
                    yield f"auth on field {ident!r} should go on the table instead"
            except SchemaObjectNotFound as e:
                yield f"{ident!r} listed in identifier list {table.identifier}, but: {e}"


@_register_validator("Identifier field with the wrong type")
def _id_type(dataset: DatasetSchema) -> Iterator[str]:
    """Identifier fields should have type integer or string."""
    for table in dataset.tables:
        for ident in table.identifier:
            try:
                field = table.get_field_by_id(ident)
                if field.type not in ["integer", "string"]:
                    yield (
                        f"identifier field {ident!r} should be a string or integer,"
                        f" is {field.type!r}"
                    )
            except SchemaObjectNotFound as e:
                yield f"{ident!r} listed in identifier list {table.identifier}, but: {e}"


@_register_validator("PostgreSQL identifier length")
def _postgres_identifier_length(dataset: DatasetSchema) -> Iterator[str]:
    """Validate inferred PostgreSQL table names for not exceeding max length.

    PostgreSQL has a maximum length for identifiers such as table names.
    We infer table names from dataset and table ids in the schemas.
    Those inferred table names should not exceed the max identifier length
    supported by PostgreSQL.
    """
    for table in dataset.get_tables(include_nested=True, include_through=True):
        db_name = table.db_name_variant(
            with_dataset_prefix=True, with_version=True, check_assert=False
        )
        # print(f"{db_name:>{MAX_TABLE_NAME_LENGTH}}")
        if (length := len(db_name)) > MAX_TABLE_NAME_LENGTH:
            excess = length - MAX_TABLE_NAME_LENGTH
            yield (
                f"Inferred PostgreSQL table name '{db_name}' is '{excess}' characters "
                f"too long. Maximum table name length is '{MAX_TABLE_NAME_LENGTH}'. Define "
                f"a `shortname`!"
            )


@_register_validator("repetitive identifiers")
def _repetitive_naming(dataset: DatasetSchema) -> Iterator[str]:
    """Identifier names should not repeat enclosing dataset/table names.

    This catches any dataset with a datasetThing table addressed by
    datasetThingIdentifier (should be dataset, thing, identifier).
    We make an exception for the case where dataset and table names are equal.
    """
    for table in dataset.tables:
        if table.id != dataset.id and table.id.startswith(dataset.id):
            yield f"table name {table.id!r} should not start with {dataset.id!r}"
        # NOTE: The code below is temporarily commented out because a lot of datasets are not
        # compliant with this rule (the precommit hook in ams-schema was
        # misconfigured, causing it to bypass checks on a lot of schemas for a long time).
        #
        # Making all datasets compliant with this is a lot of work and there is no known
        # component downstream which breaks because of violation of this rule, so until
        # we get all datasets compliant, we bypass this check.

        # for field in table.fields:
        # for prefix in [dataset.id, table.id]:
        # if field.id.startswith(prefix):
        # yield f"field name {field.id!r} should not start with {prefix!r}"


@_register_validator("identifier properties")
def _identifier_properties(dataset: DatasetSchema) -> Iterator[str]:
    """Validate that the identifier property refers to actual fields on the table definitions."""

    @dataclass
    class DerivedField:
        original: str
        derived: str

    for table in dataset.get_tables(include_nested=True):
        identifiers = set(table.identifier)
        table_fields = cast(set[str], set(map(operator.attrgetter("id"), table.fields)))
        if not identifiers.issubset(table_fields):
            missing_fields = identifiers - table_fields
            # The 'identifier' property is weird in that it is not exclusively defined in
            # terms of literally defined fields on the table. For instance, given a relation:
            #
            #     "indicatorDefinitie": {
            #       "type": "string",
            #       "relation": "bbga:indicatorenDefinities",
            #        "description": "De variabele in kwestie."
            #     }
            #
            # 'identifier' can refer to this field as 'identifierDefinitionId' (mind the
            # 'Id' postfix). Simply referring to this field (from 'identifier') as
            # 'indicatorDefinitie', eg as:
            #
            #     "identifier": ["indicatorDefinitie", "jaar", "gebiedcode15"],
            #
            #  will NOT work. It has to be postfixed with 'Id', eg:
            #
            #     "identifier": ["indicatorDefinitieId", "jaar", "gebiedcode15"],
            #
            # I think this is a bug is schema-tools, but for now I'll cover this case
            # explicitly.
            remove_id_suffix = cast(Callable[[str], str], partial(re.sub, r"(.+)Id", r"\1"))
            derived_fields = tuple(
                DerivedField(original=remove_id_suffix(f), derived=f) for f in missing_fields
            )
            for df in derived_fields:
                if df.original in table_fields:
                    missing_fields.discard(df.derived)
            if missing_fields:
                fields, have = ("fields", "have") if len(missing_fields) > 1 else ("field", "has")
                yield (
                    f"Property 'identifier' on table '{table.id}' refers to {fields} "
                    f"'{', '.join(missing_fields)}' that {have} not been defined on the "
                    "table."
                )


@_register_validator("active versions")
def _active_versions(dataset: DatasetSchema) -> Iterator[str]:
    """Validate activeVersions and table identifiers in referenced tables."""
    # The current Amsterdam Meta Schema does not allow for inline definitions of multiple
    # active tables versions. In addition :class:`DatasetSchema`'s
    # :property:`~DatasetSchema.tables` property and :meth:~DatasetSchema.get_tables` method
    # still assume that there will always be one and only one version. The part of
    # :class:`DatasetSchema` that has gained some knowledge of multiple active versions is
    # its internal representation with the addition of the :class:`TableVersions` class.
    # Hence it is the internal representation that we use for this validation.
    #
    # This obviously is a stop gap. Ideally we have a more, arguably, sensible definition of
    # multiple active version in the Amsterdam Meta Schema (eg an inline definition). When
    # we do, we are in a position to restructure our abstraction (eg :class:`DatasetSchema`,
    # etc) more definitely. And as a result can rely on those abstractions for our
    # validation instead of some internal representation.
    for table_versions in dataset.table_versions.values():
        for version in table_versions:
            try:
                # Runtime checking already happens on retrieval of the tables,
                # so this validation check only tests whether that would happen.
                table_versions[version]
            except RuntimeError as e:
                yield str(e)

        # See if default version exists
        try:
            table_versions[table_versions._default_version]
        except SchemaObjectNotFound as e:
            yield str(e)


@_register_validator("mainGeometry")
def _check_maingeometry(dataset: DatasetSchema) -> Iterator[str]:
    for table in dataset.tables:
        # We can't use table.main_geometry here, because it has a default value
        # "geometry". We can't rely on that always existing.
        main_geo = table["schema"].get("mainGeometry")
        if main_geo is None:
            # mainGeometry should exist if a geometry field exists
            # but none of the geometry fields is called "geometry"
            if table.has_geometry_fields and not any(
                field.is_geo and field.id == "geometry" for field in table.fields
            ):
                yield (
                    f"'mainGeometry' is required but not defined in table ${table.id}."
                    "This table has fields of type geometry,"
                    "but none of these fields is called 'geometry'."
                )
            continue

        try:
            field = table.get_field_by_id(main_geo)
            if not field.is_geo:
                yield f"mainGeometry = {field.id!r} is not a geometry field, type = {field.type!r}"
        except SchemaObjectNotFound as e:
            yield f"mainGeometry = {main_geo!r}, but: {e}"


@_register_validator("crs")
def _check_crs(dataset: DatasetSchema) -> Iterator[str]:
    """Check that a valid crs exists for each geometry field."""
    if dataset.data.get("crs") is None:
        for table in dataset.tables:
            if table.data.get("crs") is None:
                for field in table.fields:
                    if field.is_geo and field.crs is None:
                        yield (
                            f"No coordinate reference system defined for field {field.name}. "
                            "A crs property should exist on the field or its parent table "
                            'or parent dataset. suggestion: "EPSG:28992".'
                        )


@_register_validator("display")
def _check_display(dataset: DatasetSchema) -> Iterator[str]:
    for table in dataset.tables:
        display_field_id = table["schema"].get("display")
        if display_field_id is None:
            continue

        try:
            field = table.get_field_by_id(display_field_id)
            if field.auth != {"OPENBAAR"}:
                yield (
                    f"'auth' property on the display field: {display_field_id!r} is not allowed. "
                    " Display fields can not have an 'auth' property."
                )
        except SchemaObjectNotFound as e:
            yield f"display = {display_field_id!r}, but: {e}"


@_register_validator("property formats")
def _property_formats(dataset: DatasetSchema) -> Iterator[str]:
    """Properties should have a valid "format", or none at all."""
    # TODO Should we be validating these in the meta-schema instead of here?
    ALLOWED = {
        # Default value for DatasetFieldSchema.format.
        None,
        # Listed in the schema spec.
        "date",
        "date-time",
        "duration",
        "email",
        "hostname",
        "idn-email",
        "idn-hostname",
        "ipv4",
        "ipv6",
        "iri",
        "iri-reference",
        "time",
        "uri",
        "uri-reference",
        # XXX In actual use, not sure what it's supposed to mean.
        "summary",
    }

    for table in dataset.tables:
        for field in table.fields:
            if field.type == "str" and field.format not in ALLOWED:
                yield f"Format {field.format!r} not allowed, must be one of {ALLOWED!r}"


@_register_validator("auth across relations")
def _relation_auth(dataset: DatasetSchema) -> Iterator[str]:
    """Relation fields should have at least the auth scopes of the field they refer to."""
    for table in dataset.tables:
        for field in table.get_fields(include_subfields=True):
            our_auth = table.dataset.auth | table.auth | field.auth

            rel_table = field.related_table
            if not rel_table:
                continue

            if (
                not our_auth.issuperset(rel_table.dataset.auth)
                or not our_auth.issuperset(rel_table.auth)
                or not all(
                    our_auth.issuperset(rel_table.get_field_by_id(f).auth)
                    for f in (field.related_field_ids or [])
                )
            ):
                scopes = set(rel_table.dataset.auth) | set(rel_table.auth)
                for f in field.related_field_ids:
                    scopes |= rel_table.get_field_by_id(f).auth
                scopes.remove("OPENBAAR")  # Not very interesting.
                yield f"{table.id}.{field.id} requires scopes {sorted(scopes)}"


@_register_validator("reasons non public exists")
def _reasons_non_public_exists(dataset: DatasetSchema) -> Iterator[str]:
    """A ReasonsNonPublic field should be present on the highest non-public scope.

    For non-public fields in a public table, reasonsNonPublic can be set on table
    level. This is less verbose when all the non-public fields have the same
    reason for being non-public.

    """
    if dataset.auth == {"OPENBAAR"}:
        for table in dataset.tables:
            if table.data.get("reasonsNonPublic") is None:
                if table.auth == {"OPENBAAR"}:
                    for field in table.fields:
                        if (
                            field.auth != {"OPENBAAR"}
                            and field.data.get("reasonsNonPublic") is None
                        ):
                            yield (
                                f"Non-public field {field.id} or it's parent table "
                                "should have a 'reasonsNonPublic' property."
                            )
                else:
                    yield (
                        f"Non-public table {table.id} should have a 'reasonsNonPublic' property."
                    )
    elif dataset.data.get("reasonsNonPublic") is None:
        yield (f"Non-public dataset {dataset.id} should have a 'reasonsNonPublic' property.")


@_register_validator("reasons non public value")
def _reasons_non_public_value(dataset: DatasetSchema) -> Iterator[str]:
    """A reasonsNonPublic field in a published dataset should not contain a placeholder."""
    placeholder_value = "nader te bepalen"
    if dataset.data.get("status") != "beschikbaar":
        return
    if placeholder_value in dataset.data.get("reasonsNonPublic", []):
        yield (
            f"Placeholder value '{placeholder_value}' not allowed in "
            f"ReasonsNonPublic property of dataset {dataset.id}."
        )
    for table in dataset.tables:
        if placeholder_value in table.data.get("reasonsNonPublic", []):
            yield (
                f"Placeholder value '{placeholder_value}' not allowed "
                f"ReasonsNonPublic property of table {table.id}."
            )
        for field in table.fields:
            if placeholder_value in field.data.get("reasonsNonPublic", []):
                yield (
                    f"Placeholder value '{placeholder_value}' not allowed "
                    f"ReasonsNonPublic property of field {field.id}."
                )


@_register_validator("schema ref")
def _check_schema_ref(dataset: DatasetSchema) -> Iterator[str]:
    """Check that $ref field for all tables has correct hostname."""
    for table in dataset.tables:
        fragments = urlparse(table["schema"]["properties"]["schema"]["$ref"])
        if fragments.hostname != "schemas.data.amsterdam.nl" or fragments.scheme != "https":
            yield (
                f"Incorrect `$ref` for {table.id}. Value should be "
                f"`https://schemas.data.amsterdam.nl`"
            )


@_register_validator("defaultVersion")
def _check_default_version(dataset: DatasetSchema) -> Iterator[str]:
    """Check that defaultVersion field for a dataset matches the only 'beschikbaar' version."""
    enabled_versions = []
    for version_number, version in dataset.versions.items():
        if version.status == DatasetSchema.Status.beschikbaar:
            enabled_versions.append(version_number)
            if version_number != dataset.default_version:
                yield (
                    f"Default version {dataset.default_version} does not match enabled "
                    f"version {version_number}"
                )


@_register_validator("production version tables")
def _check_production_version_tables(dataset: DatasetSchema) -> Iterator[str]:
    """Check that a production version (>= v1) contains no v0 tables."""
    non_production_versions = ["v0"]
    version_regex = r"(v?(?P<major>\d+)(?:\.(?P<minor>\d+)(?:\.(?P<patch>\d+))?)?)"
    for version_number, version in dataset.versions.items():
        if version_number in non_production_versions:
            continue

        for table in version.data.get("tables"):
            # Don't validate inline tables, to allow backwards compatibility
            if "$ref" in table:
                table_version = re.search(version_regex, table["$ref"]).group(0)
                if table_version in non_production_versions:
                    yield (
                        f"Dataset version ({version_number}) cannot contain non-production "
                        f"table [{table['$ref']}]"
                    )


@_register_validator("stable version experimental tables")
def _check_production_version_experimental_tables(dataset: DatasetSchema) -> Iterator[str]:
    """Check that a stable dataset version contains no experimental tables."""
    for version_number, version in dataset.versions.items():
        if version.lifecycle_status == DatasetVersionSchema.LifecycleStatus.experimental:
            continue

        for table in version.tables:
            # Don't validate inline tables, to allow backwards compatibility
            if table.lifecycle_status == DatasetTableSchema.LifecycleStatus.experimental:
                yield (
                    f"Stable dataset {dataset.id} ({version_number}) cannot have tables with "
                    f"lifecycleStatus of 'experimental'."
                )


@_register_validator("lifecycle status")
def _check_lifecycle_status(dataset: DatasetSchema) -> Iterator[str]:
    """
    Check that a non production version (< v1) contains can't have a
    lifecycle status of 'stable'.
    """
    non_production_versions = ["v0"]
    for version_number, version in dataset.versions.items():
        if (
            version_number in non_production_versions
            and version.lifecycle_status == DatasetVersionSchema.LifecycleStatus.stable
        ):
            yield (
                f"Dataset version ({version_number}) cannot have a lifecycleStatus of "
                f"'stable' while being a non-production version."
            )


PROPERTIES_INTRODUCING_BREAKING_CHANGES = ["type", "$ref", "format", "relation", "enum"]


def validate_table(
    previous_fields: dict,
    current_fields: dict,
    parent_field: str | None = None,
    object_path: str | None = None,
) -> list[str]:
    """Validates that the current version of the table does not introduce breaking changes.

    We check:
    1. whether a field has been deleted
    2. whether any of the properties in `PROPERTIES_INTRODUCING_BREAKING_CHANGES` has changed

    This works recursively for array items and object properties.

    NB. This is not a registered validator that has been registered, as here we have to
    compare fields from two versions of the same table.
    """
    table_errors = []
    current_object_path = object_path
    for previous_field_name, previous_field in previous_fields.items():
        column_name = parent_field or previous_field_name

        # check if field is deleted
        if previous_field_name not in current_fields:
            if parent_field:
                table_errors.append(
                    f"Property {object_path}.{previous_field_name} would be "
                    f"deleted from column {column_name}."
                )
            else:
                table_errors.append(f"Column {column_name} would be deleted.")
            continue

        current_field = current_fields[previous_field_name]

        # check breaking changes in field properties
        for prop in PROPERTIES_INTRODUCING_BREAKING_CHANGES:
            if previous_field.get(prop) != current_field.get(prop):
                if parent_field:
                    current_object_path = (
                        f"{object_path}.{previous_field_name}"
                        if object_path
                        else previous_field_name
                    )
                prop_name = f"{current_object_path}.{prop}" if current_object_path else prop
                table_errors.append(f"Column {column_name} would change {prop_name}.")

        # recursively check array items
        if previous_field.get("type") == "array":
            previous_array_items = {"items": previous_field.get("items")}
            current_array_items = {"items": current_field.get("items")}
            table_errors.extend(
                validate_table(previous_array_items, current_array_items, column_name)
            )

        # recursively check object properties
        if previous_field.get("type") == "object":
            previous_object_properties = previous_field.get("properties")
            current_object_properties = current_field.get("properties")
            if previous_field_name != column_name:
                object_path = (
                    previous_field_name
                    if object_path is None
                    else f"{object_path}.{previous_field_name}"
                )

            table_errors.extend(
                validate_table(
                    previous_object_properties,
                    current_object_properties,
                    column_name,
                    object_path,
                )
            )

    return table_errors



# Generated by Django 4.1.7 on 2024-02-29 11:34
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("datasets", "0017_dataset_enable_export"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="id_field",
            field=models.CharField(default="id", max_length=50),
        ),
    ]



from __future__ import annotations

from collections.abc import Iterable
from typing import Any

from django.conf import settings
from django.core.management import BaseCommand, CommandError, CommandParser

from schematools.contrib.django.models import Dataset


class BaseDatasetCommand(BaseCommand):
    """A management command that works on the list of imported datasets."""

    def add_arguments(self, parser: CommandParser) -> None:
        """Provide default arguments to pass dataset names to this management command."""
        parser.add_argument(
            "dataset", nargs="*", help="Datasets to use. Takes precedent over --datasets-list"
        )
        parser.add_argument(
            "--datasets-list",
            nargs="*",
            default=settings.DATASETS_LIST,
            help=f"Datasets to use (default: {settings.DATASETS_LIST})",
        )
        parser.add_argument(
            "-x",
            "--exclude",
            "--datasets-exclude",
            dest="datasets_exclude",
            nargs="*",
            default=settings.DATASETS_EXCLUDE,
            help="Datasets that need to be skipped. (default: {settings.DATASETS_EXCLUDE})",
        )

    def get_datasets(
        self, options: dict[str, Any], enable_db=None, default_all=False
    ) -> Iterable[Dataset]:
        """Provide the datasets based on the command options"""
        # Provide backwards compatibility for the positional argument datasets
        options["datasets_list"] = (
            options["dataset"] if options["dataset"] else options["datasets_list"]
        )

        if not options["datasets_list"] and not options["datasets_exclude"] and not default_all:
            raise CommandError(
                "Provide at least a dataset using --datasets-list, "
                "or use the --datasets-exclude option."
            )
        queryset = Dataset.objects.all()
        if enable_db is not None:
            queryset = queryset.filter(enable_db=enable_db)

        datasets = {ds.name for ds in queryset}
        if options["datasets_list"] is not None:
            datasets = set(options["datasets_list"])

        if options["datasets_exclude"] is not None:
            datasets = datasets - set(options["datasets_exclude"])

        queryset = queryset.filter(name__in=datasets)
        if invalid_names := datasets - {ds.name for ds in queryset}:
            raise CommandError(f"Datasets not found: {', '.join(sorted(invalid_names))}")

        return queryset



from __future__ import annotations

from schematools.types import DatasetSchema


class ProvenanceIteration:
    """Extract provenance elements in DatasetSchema (json)
    so it can be used for automatic processing"""

    final_dic = {}
    final_dic_all_columns = {}
    temp_dic_tables = {}
    temp_dic_columns = {}
    number_of_tables = 0

    def __init__(self, dataschema: DatasetSchema):
        """Trigger processing"""

        self.set_number_of_tables(dataschema)
        self.set_dataset_for_final_listing(dataschema)

    def set_number_of_tables(self, dataschema: DatasetSchema):
        """Retrieving the number of tables in datasetschema for looping purpose"""
        for item in dataschema:
            if item == "tables":
                self.number_of_tables = len(dataschema[item])
                return self.number_of_tables
        return None

    def set_dataset_for_final_listing(self, dataschema: DatasetSchema):
        """Setting dataset level to add later als wrapper"""

        if isinstance(dataschema, dict):

            # At first make root branch dataset
            # and take id field as value for dataschema attribute
            self.final_dic["dataset"] = dataschema["id"]
            # Add provenance element on dataset level
            self.final_dic["provenance"] = dataschema.get("provenance", "na")
            # and make tables branch to hold the tables provenance data
            self.final_dic["tables"] = []

            for item in dataschema:
                if item == "tables":
                    self.get_provenance_per_table(dataschema[item])

    def get_provenance_per_table(self, dicts):
        """Calling the processing to extract the provenance data per table"""

        # loop trough all existing tables and proces data for provenance
        # on each and add them each to the final_dict outcome
        for d in dicts[: self.number_of_tables]:
            self.temp_dic_tables["table"] = d["id"]
            self.temp_dic_tables["provenance"] = d.get("provenance", "na")
            self.temp_dic_tables["properties"] = []
            self.get_table_columns(d["schema"]["properties"])
            # add table columns (within element properties) to table
            self.temp_dic_tables["properties"].append(dict(self.temp_dic_columns))
            # add table result (within element tables) in final result
            self.final_dic["tables"].append(self.temp_dic_tables)
            # clean up temp_dic* to enforce a new object for next table
            self.temp_dic_tables = {}
            self.temp_dic_columns = {}

    def get_table_columns(self, dictionary):
        """Extrating the columns and if provenance then adding its value,
        resulting in a dictionary of source name : column name"""

        done = set()
        self.temp_dic_columns = {}
        for column in dictionary:
            if "provenance" in dictionary[column]:
                done.add(column)
                self.temp_dic_columns[column] = dictionary[column]["provenance"]

            # Add the columns to the dictionary that have no provenance
            # and exclude schema as it will never be added in the database
            if column not in done and column != "schema":
                self.temp_dic_columns[column] = column



# Generated by Django 3.0.7 on 2020-09-21 14:54
from __future__ import annotations

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0007_datasettable_is_temporal"),
    ]

    operations = [
        migrations.CreateModel(
            name="Profile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("scopes", models.CharField(max_length=255)),
                (
                    "schema_data",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        verbose_name="Amsterdam Schema Contents"
                    ),
                ),
            ],
        ),
    ]



"""The Django models for Amsterdam Schema data.

When models are generated with :func:`~schematools.contrib.django.factories.model_factory`,
they all inherit from :class:`~schematools.contrib.django.models.DynamicModel` to have
a common interface.

When the schema data is imported, the models
:class:`~schematools.contrib.django.models.Dataset`,
:class:`~schematools.contrib.django.models.DatasetTable`,
:class:`~schematools.contrib.django.models.DatasetField` and
:class:`~schematools.contrib.django.models.Profile` are all filled.
"""

from __future__ import annotations

import json
import logging
import re
from typing import TypeVar

from django.conf import settings
from django.db import models, transaction
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from schematools.exceptions import DatasetFieldNotFound
from schematools.loaders import CachedSchemaLoader
from schematools.naming import to_snake_case
from schematools.types import (
    DatasetFieldSchema,
    DatasetSchema,
    DatasetTableSchema,
    ProfileSchema,
    SemVer,
)

from . import managers
from .validators import URLPathValidator, validate_json

logger = logging.getLogger(__name__)


class DynamicModel(models.Model):
    """Base class to tag and detect dynamically generated models."""

    #: Overwritten by subclasses / factory
    CREATION_COUNTER = None
    _dataset: Dataset = None  # type: ignore[assignment]
    _dataset_schema: DatasetSchema = None  # type: ignore[assignment]
    _table_schema: DatasetTableSchema = None  # type: ignore[assignment]
    _display_field = None
    _is_temporal = None

    class Meta:
        abstract = True

    def __str__(self) -> str:
        if self._display_field:
            return str(getattr(self, self._display_field))
        else:
            # this will not be shown in the dso-api view
            # and will be omitted when display field is empty or not present
            return f"(no title: {self._meta.object_name} #{self.pk})"

    # These classmethods could have been a 'classproperty',
    # but this ensures the names don't conflict with fields from the schema.
    @classmethod
    def get_dataset(cls) -> Dataset:
        """Give access to the original dataset that this models is part of."""
        return cls._dataset

    @classmethod
    def get_dataset_id(cls) -> str:
        """Give access to the original dataset ID that this model is part of."""
        return cls._dataset_schema.id

    @classmethod
    def get_dataset_path(cls) -> str:
        """Give access to the api path this dataset should be published on."""
        return cls._dataset.path

    @classmethod
    def get_dataset_schema(cls) -> DatasetSchema:
        """Give access to the original dataset schema that this model is a part of."""
        return cls._dataset_schema

    @classmethod
    def table_schema(cls) -> DatasetTableSchema:
        """Give access to the original table_schema that this model implements."""
        return cls._table_schema

    @classmethod
    def get_field_schema(
        cls, model_field: models.Field | models.ForeignObjectRel
    ) -> DatasetFieldSchema:
        """Provide access to the underlying amsterdam schema field that created the model field."""
        if isinstance(model_field, models.ForeignObjectRel):
            # When Django auto-creates the related field, it doesn't have `field_schema`,
            # but it can be resolved by looking up the original forward relation.
            model_field = model_field.remote_field

        try:
            # This internal property is assigned by model_factory()
            return model_field.field_schema
        except AttributeError:
            # Easier to ask for forgiveness than permission;
            if not issubclass(model_field.model, DynamicModel):
                raise ValueError(
                    "get_field_schema() is only usable on fields from on DynamicModel instances."
                ) from None

            if model_field.auto_created:
                raise ValueError(
                    "get_field_schema() can't be used on"
                    f" '{model_field.model._meta.model_name}.{model_field.name}',"
                    f" because that is an auto-created field."
                ) from None
            raise

    @classmethod
    def get_table_id(cls) -> str:
        """Give access to the table name"""
        return cls._table_schema.id

    @classmethod
    def has_parent_table(cls) -> bool:
        """Check if table is sub table for another table."""
        return cls._table_schema.has_parent_table

    @classmethod
    def has_display_field(cls) -> bool:
        """Tell whether a display field is configured."""
        return cls._display_field is not None

    @classmethod
    def get_display_field(cls) -> str | None:
        """Return the name of the display field, for usage by Django models."""
        return cls._display_field

    @classmethod
    def is_temporal(cls) -> bool:
        """Indicates if this model has temporary characteristics."""
        return cls._is_temporal


M = TypeVar("M", bound=DynamicModel)


class Dataset(models.Model):
    """A registry of all available datasets that are uploaded in the API server.

    Each model holds the contents of an "Amsterdam Schema",
    that contains multiple tables.
    """

    name = models.CharField(_("Name"), unique=True, max_length=50)
    schema_data = models.TextField(_("Amsterdam Schema Contents"), validators=[validate_json])
    view_data = models.TextField(_("View SQL"), blank=True, null=True)
    version = models.CharField(_("Schema Version"), blank=True, null=True, max_length=250)
    is_default_version = models.BooleanField(_("Default version"), default=False)

    # Settings for publishing the schema:
    enable_api = models.BooleanField(default=True)
    enable_db = models.BooleanField(default=True)
    enable_export = models.BooleanField(default=False)
    path = models.TextField(unique=True, blank=False, validators=[URLPathValidator()])
    auth = models.CharField(_("Authorization"), blank=True, null=True, max_length=250)
    ordering = models.IntegerField(_("Ordering"), default=1)

    objects = managers.DatasetQuerySet.as_manager()

    class Meta:
        ordering = ("ordering", "name")
        verbose_name = _("Dataset")
        verbose_name_plural = _("Datasets")

    def __str__(self):
        return self.name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dataset_collection = None

        # The check makes sure that deferred fields are not checked for changes,
        # nor that creating the model
        self._old_schema_data = (
            self.schema_data if "schema_data" in self.__dict__ and not self._state.adding else None
        )

    def save(self, *args, **kwargs):
        """Perform a final data validation check, and additional updates."""
        if self.schema_data_changed() and (self.schema_data or not self._state.adding):
            self.__dict__.pop("schema", None)  # clear cached property
            # The extra "and" above avoids the transaction savepoint for an empty dataset.
            # Ensure both changes are saved together
            with transaction.atomic():
                super().save(*args, **kwargs)
                self.save_schema_tables()
        else:
            super().save(*args, **kwargs)

    save.alters_data = True

    @classmethod
    def name_from_schema(cls, schema: DatasetSchema) -> str:
        """Generate dataset name from schema"""
        name = to_snake_case(schema.id)
        if schema.version and not schema.is_default_version:
            name = to_snake_case(f"{name}_{schema.version}")
        return name

    @classmethod
    def create_for_schema(
        cls, schema: DatasetSchema, path: str | None = None, enable_db: bool = True
    ) -> Dataset:
        """Create the schema based on the Amsterdam Schema JSON input"""
        name = cls.name_from_schema(schema)
        if path is None:
            path = name
        obj = cls(
            name=name,
            schema_data=schema.json(
                inline_tables=True, inline_publishers=True, inline_scopes=True
            ),
            view_data=schema.get_view_sql(),
            auth=" ".join(schema.auth),
            path=path,
            version=schema.version,
            is_default_version=schema.is_default_version,
            enable_api=cls.has_api_enabled(schema),
            enable_db=enable_db,
        )
        obj._dataset_collection = schema.loader  # retain collection on saving
        obj.save()
        obj.__dict__["schema"] = schema  # Avoid serializing/deserializing the schema data
        return obj

    def save_path(self, path: str) -> bool:
        """Update this model with new path"""
        if path_changed := self.path != path:
            self.path = path
            self.save(update_fields=["path"])
        return path_changed

    def save_for_schema(self, schema: DatasetSchema):
        """Update this model with schema data"""
        self.schema_data = schema.json(
            inline_tables=True, inline_publishers=True, inline_scopes=True
        )
        self.view_data = schema.get_view_sql()
        self.auth = " ".join(schema.auth)
        self.enable_api = Dataset.has_api_enabled(schema)
        self._dataset_collection = schema.loader  # retain collection on saving

        changed = self.schema_data_changed()
        if changed:
            self.save(
                update_fields=[
                    "schema_data",
                    "view_data",
                    "auth",
                    "is_default_version",
                    "enable_api",
                ]
            )

        self.__dict__["schema"] = schema  # Avoid serializing/deserializing the schema data
        return changed

    def save_schema_tables(self):
        """Expose the schema data to the DatasetTable.
        This allows other projects (e.g. geosearch) to process our dynamic tables.
        """
        if not self.schema_data:
            # no schema stored -> no tables
            if self._old_schema_data:
                self.tables.all().delete()
            return

        new_definitions = {
            to_snake_case(t.id): t for t in self.schema.get_tables(include_nested=True)
        }
        new_names = set(new_definitions.keys())
        existing_models = {t.name: t for t in self.tables.all()}
        existing_names = set(existing_models.keys())

        # Create models for newly added tables
        for added_name in new_names - existing_names:
            table = new_definitions[added_name]
            DatasetTable.create_for_schema(self, table)

        # Update models for updated tables
        for changed_name in existing_names & new_names:
            table = new_definitions[changed_name]
            existing_models[changed_name].save_for_schema(table)

        # Remove tables that are no longer part of the schema.
        for removed_name in existing_names - new_names:
            existing_models[removed_name].delete()

    save_schema_tables.alters_data = True

    @cached_property
    def schema(self) -> DatasetSchema:
        """Provide access to the schema data."""
        # The _dataset_collection value is filled by the queryset,
        # so any object that is fetched by same the queryset uses the same shared cache.
        return self.get_schema(dataset_collection=self._dataset_collection)

    def get_schema(self, dataset_collection: CachedSchemaLoader) -> DatasetSchema:
        """Extract the schema data from this model, and connect it with a dataset collection."""
        if not self.schema_data:
            raise RuntimeError("Dataset.schema_data is empty")

        return DatasetSchema.from_dict(
            json.loads(self.schema_data),
            dataset_collection=dataset_collection,
        )

    @cached_property
    def has_geometry_fields(self) -> bool:
        return any(table.has_geometry_fields for table in self.schema.tables)

    @classmethod
    def has_api_enabled(cls, schema: DatasetSchema) -> bool:
        dataset_status = schema.status
        if dataset_status == DatasetSchema.Status.beschikbaar:
            return True
        elif dataset_status == DatasetSchema.Status.niet_beschikbaar:
            return False

        raise ValueError(
            f"Cannot determine whether to enable REST api based on given status: {dataset_status}"
        )

    def schema_data_changed(self):
        """Check whether the schema_data attribute changed"""
        return (
            "schema_data" in self.__dict__  # this checks for deferred attributes
            and self.schema_data != self._old_schema_data
        )

    def create_models(
        self, base_app_name: str | None = None, base_model: type[M] = DynamicModel
    ) -> list[type[M]]:
        """Extract the models found in the schema"""
        from schematools.contrib.django.factories import schema_models_factory

        if not self.enable_db:
            return []
        else:
            return schema_models_factory(self, base_app_name=base_app_name, base_model=base_model)


class DatasetTable(models.Model):
    """Exposed metadata per schema.

    This table can be read by the 'geosearch' project to locate all our tables and data sources.
    """

    dataset = models.ForeignKey(Dataset, on_delete=models.CASCADE, related_name="tables")
    name = models.CharField(max_length=100)
    version = models.TextField(default=SemVer("1.0.0"))

    # Exposed metadata from the jsonschema, so other apps (e.g. geosearch) can query these
    auth = models.CharField(max_length=250, blank=True, null=True)
    enable_export = models.BooleanField(default=False)
    enable_geosearch = models.BooleanField(default=True)
    db_table = models.CharField(max_length=100)
    display_field = models.CharField(max_length=50, null=True, blank=True)
    geometry_field = models.CharField(max_length=50, null=True, blank=True)
    geometry_field_type = models.CharField(max_length=50, null=True, blank=True)
    is_temporal = models.BooleanField(null=False, blank=False, default=False)
    id_field = models.CharField(max_length=50, blank=False, default="id")

    class Meta:
        ordering = ("name",)
        verbose_name = _("Dataset Table")
        verbose_name_plural = _("Dataset Tables")
        unique_together = [
            ("dataset", "name"),
        ]

    def __str__(self):
        return self.name

    @classmethod
    def _get_geometry_field(cls, table_schema):
        try:
            field = table_schema.main_geometry_field
        except DatasetFieldNotFound:
            # fallback to first GeoJSON field as geometry field.
            field = next((f for f in table_schema.fields if f.is_geo), None)
            if field is None:
                return None, None

        match = re.search(r"schema\/(?P<schema>\w+)\.json", field.type)
        geo_type = match.group("schema") if match is not None else None
        return field.db_name, geo_type

    @classmethod
    def _get_id_field(cls, table_schema):
        """Gets the id_field, this is the PK that is needed by Django and geosearch.

        An `id` field will be added to the schema for tables with a compound key.
        """
        identifier = table_schema.identifier
        return to_snake_case(identifier[0]) if len(identifier) == 1 else "id"

    @classmethod
    def create_for_schema(cls, dataset: Dataset, table_schema: DatasetTableSchema) -> DatasetTable:
        """Create a DatasetTable object based on the Amsterdam Schema table spec.

        (The table spec contains a JSON-schema for all fields).
        """
        instance = cls(dataset=dataset)
        instance.save_for_schema(table_schema)
        return instance

    def save_for_schema(self, table_schema: DatasetTableSchema):
        """Save changes to the dataset table schema."""
        display_field = table_schema.display_field
        self.name = to_snake_case(table_schema.id)
        self.db_table = table_schema.db_name
        self.auth = " ".join(table_schema.auth)
        self.display_field = display_field.db_name if display_field is not None else None
        self.geometry_field, self.geometry_field_type = self._get_geometry_field(table_schema)
        self.is_temporal = table_schema.is_temporal
        self.enable_geosearch = (
            table_schema.dataset.id not in settings.AMSTERDAM_SCHEMA["geosearch_disabled_datasets"]
        )
        self.id_field = self._get_id_field(table_schema)

        is_creation = not self._state.adding
        self.save()

        new_definitions = {f.python_name: f for f in table_schema.fields}
        new_names = set(new_definitions.keys())
        existing_fields = {f.name: f for f in self.fields.all()} if is_creation else {}
        existing_names = set(existing_fields.keys())

        # Create new fields
        for added_name in new_names - existing_names:
            DatasetField.create_for_schema(self, field=new_definitions[added_name])

        # Update existing fields
        for changed_name in existing_names & new_names:
            field = new_definitions[changed_name]
            existing_fields[changed_name].save_for_schema(field=field)

        for removed_name in existing_names - new_names:
            existing_fields[removed_name].delete()


class DatasetField(models.Model):
    """Exposed metadata per field."""

    table = models.ForeignKey(DatasetTable, on_delete=models.CASCADE, related_name="fields")
    name = models.CharField(max_length=100)

    # Exposed metadata from the jsonschema, so other utils can query these
    auth = models.CharField(max_length=250, blank=True, null=True)

    class Meta:
        ordering = ("name",)
        verbose_name = _("Dataset Field")
        verbose_name_plural = _("Dataset Fields")
        unique_together = [
            ("table", "name"),
        ]

    def __str__(self):
        return self.name

    @classmethod
    def create_for_schema(cls, table: DatasetTable, field: DatasetFieldSchema) -> DatasetField:
        """Create a DatasetField object based on the Amsterdam Schema field spec."""
        instance = cls(table=table)
        instance.save_for_schema(field)
        return instance

    def save_for_schema(self, field: DatasetFieldSchema):
        """Update the field with the provided schema data."""
        self.name = to_snake_case(field.id)
        self.auth = " ".join(field.auth)
        self.save()


class Profile(models.Model):
    """User Profile."""

    name = models.CharField(max_length=100)
    scopes = models.CharField(max_length=255)
    schema_data = models.TextField(_("Amsterdam Schema Contents"), validators=[validate_json])

    def __str__(self):
        return self.name

    @cached_property
    def schema(self) -> ProfileSchema:
        """Provide access to the schema data"""
        if not self.schema_data:
            raise RuntimeError("Profile.schema_data is empty")

        return ProfileSchema.from_dict(json.loads(self.schema_data))

    def get_scopes(self):
        """The auth scopes for this profile"""
        return set(json.loads(self.scopes))

    @classmethod
    def create_for_schema(cls, profile_schema: ProfileSchema) -> Profile:
        """Create Profile object based on the Amsterdam Schema profile spec."""
        instance = cls()
        instance.save_for_schema(profile_schema)
        return instance

    def save_for_schema(self, profile_schema: ProfileSchema) -> Profile:
        self.name = profile_schema.name
        self.scopes = json.dumps(sorted(profile_schema.scopes))
        self.schema_data = profile_schema.json()
        self.save()
        return self


class LooseRelationField(models.ForeignKey):
    """A relation that points to one part of the composite key.

    This only points to the first field of the relation, and the second field
    (e.g. volgnummer/beginGeldigheid) is determined at runtime using query filtering.
    Without such filter, traversing the relation will produce multiple results.

    This setup is typically used for temporal relationships, where the foreign table uses compound
    keys as identifier. This field type references the first field of that compound relationship.

    At construction, the "to_field" will be set pointing to the correct identifier field.
    """

    # Disable the unique check made against the target field.
    # When it's part of a composite key, this can't always be unique by itself.
    # However, we can still have a db constraint on it.
    requires_unique_target = False

    def __init__(self, *args, **kwargs):
        if not kwargs.get("to_field"):
            raise ValueError("to_field must be provided.")
        super().__init__(*args, **kwargs)


class CompositeForeignKeyField(models.ForeignKey):
    """A composite key, based on multiple fields on the target table.

    Note this class is currently backed by a database column
    that holds a concatenated value in the database.
    So this acts as a tagging class only to reveal what the field type is.
    Ideally, this would have to change to an actual compound key field
    like Django's ``GenericForeignKey`` is. Also note that
    Django's ForeignObject has a ``to_fields`` attribute that supports
    referencing multiple fields.
    """

    def __init__(self, *args, to_fields, **kwargs):
        super().__init__(*args, **kwargs)

        # Currently, only used as descriptive data for analysis.
        # Called as _to_fields to avoid overriding the base class attribute.
        self._to_fields = to_fields

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["to_fields"] = self._to_fields
        return name, path, args, kwargs


class LooseRelationManyToManyField(models.ManyToManyField):
    pass



from __future__ import annotations

from typing import Any

from schematools.contrib.django.faker.relate import relate_datasets

from . import BaseDatasetCommand


class Command(BaseDatasetCommand):  # noqa: D101
    help = """Relate mock records.

    When mock data is created, the relations are filled with `null` values.
    Using this command, the relations can be added to the records.
    """  # noqa: A003
    requires_system_checks = []

    def handle(self, *args: list[Any], **options: dict[str, Any]) -> None:  # noqa: D102
        datasets = self.get_datasets(options, enable_db=True)
        relate_datasets(*datasets)



# Generated by Django 4.2.17 on 2025-01-29 10:47
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0018_datasettable_id_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="enable_export",
            field=models.BooleanField(default=False),
        ),
    ]
